using System;
using System.Collections;

namespace Cosmology
{

    /// <summary>
    /// Class that computes different distance measures and other quatities for any cosmological parameters.
    /// Current Version
    /// ID:         $Id: CosmologicalFunctions.cs,v 1.0 2008/06/20 10:59:55  mtaghiza Exp $
    /// Revision:   $Revision: 1.0 $
    /// Date:       $Date: 2008/06/20 10:59:55 $
    /// </summary>
    public partial class CosmologicalFunctions
    {
        // References:
        // arXiv:astro-ph/9905116v4 16 Dec 2000. ("Distance measures in cosmology",David Hogg)
        // Mon. Not. R. Astron. Soc. 319, 557-561 (2000). (On the Cosmological distance and redshift between any two objects", J. Liske) 
        // see also http://lambda.gsfc.nasa.gov/product/map/dr3/pub_papers/fiveyear/cosmology/wmap_5yr_cosmo.pdf

        const double c = 299792.458;// speed of light in Km/s (constant)
        public double w_0, k, OmegaM_0, OmegaK_0, OmegaL_0, OmegaR_0, OmegaT_0, h_0, HubbleParameter_0, HubbleDistance_0, HubbleTime_0, PhiStar, MStar, Alpha;// note that OmegaK_0=1-OmegaT=1-OmegaL_0-OmegaM_0-OmegaR_0, OmegaT=rho/rho_critical
        public double zAtDaMax;
        public double DaMax;
        public double z, D_c, D_m, D_a, D_l, ComovingVolume, LOOKbackTIME, AGEofUNIVERSE, TotalAGEofUNIVERSE;// to be instatiated when calling Set...(...)

        public double TOL_Romberg; // tolerances in the Romberg integration routine
        public int NMIN_Romberg;// tolerances in the Romberg integration routine
        public double TOL_BracketingWithNeville;//fractional tolerance in the BracketingWithNelville root finding routine.
        public int MAXIT_BracketingWithNeville;// maximum number of evaluations in the BracketingWithNelville root finding routine.
        public int NumVertices_BracketingWithNeville;
        public double ExpansionFactor_BracketingWithNeville;// to be used in the BracketingWithNeville routine.

        const double zMiddle = 20;//redshift to be used in the calculations of Time as a pivot time in the integration limits of the time integrals.
        const double z_Max = double.MaxValue;// 1.7976931348623e308;// 1.79769313486232e308 
        const double DeltaZIntegration_Dc = 2.0;//seed redshift to be used for calculating the line of sight comoving distance out of the integral.
        const int SignificantPlaces = 15;

        public const double Redshift0_Bracketing = 0.0;
        public const double Redshift1_Bracketing = 1.0;

        private NumUtils.Integration.Romberg ROMBERG;
        private NumUtils.RootFinding.BracketingWithNelville BWN;
        private NumUtils.RootFinding.BracketingWithNewtonRaphson BWNR;
        private Mathematics.MathFunctions MATH = new Mathematics.MathFunctions();

        private int ChooseTimeFunction;
        private double MaxTime = 0;

        public CosmologicalFunctions()
        {
            ResetCosmoParameters(0.3, 0.7, 0.0, -1.0, 0.7);
            ResetNumUtils();
        }

        public CosmologicalFunctions(double OmegaM_0, double OmegaL_0, double OmegaR_0, double w_0, double h_0)
        {
            ResetCosmoParameters(OmegaM_0, OmegaL_0, OmegaR_0, w_0, h_0);
            ResetNumUtils();
        }

        public CosmologicalFunctions(double OmegaM_0, double OmegaL_0, double OmegaR_0, double w_0, double h_0, double TOL_Romberg, double TOL_BracketingWithNeville)
        {
            ResetCosmoParameters(OmegaM_0, OmegaL_0, OmegaR_0, w_0, h_0);
            ResetNumUtils(TOL_Romberg, 5, TOL_BracketingWithNeville, 10, 5, 5000);
        }

        public CosmologicalFunctions(double OmegaM_0, double OmegaL_0, double OmegaR_0, double w_0, double h_0, double TOL_Romberg, int NMIN_Romberg, double TOL_BracketingWithNeville, double ExpansionFactor_BracketingWithNeville, int NumVertices_BracketingWithNeville, int MAXIT_BracketingWithNeville)
        {
            ResetCosmoParameters(OmegaM_0, OmegaL_0, OmegaR_0, w_0, h_0);
            ResetNumUtils(TOL_Romberg, NMIN_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
        }

        public void ResetNumUtils()
        {
            this.TOL_Romberg = 1e-9;
            this.NMIN_Romberg = 5;
            this.TOL_BracketingWithNeville = 1e-9;
            this.ExpansionFactor_BracketingWithNeville = 10;
            this.NumVertices_BracketingWithNeville = 5; 
            this.MAXIT_BracketingWithNeville = 5000;
            this.ROMBERG = new NumUtils.Integration.Romberg(NMIN_Romberg, TOL_Romberg);
            this.BWN = new NumUtils.RootFinding.BracketingWithNelville(this.ExpansionFactor_BracketingWithNeville, this.NumVertices_BracketingWithNeville, this.TOL_BracketingWithNeville, this.MAXIT_BracketingWithNeville);
            this.BWNR = new NumUtils.RootFinding.BracketingWithNewtonRaphson(this.ExpansionFactor_BracketingWithNeville, this.NumVertices_BracketingWithNeville, this.TOL_BracketingWithNeville, this.MAXIT_BracketingWithNeville);
        }

        public void ResetNumUtils(double TOL_Romberg, int NMIN_Romberg, double TOL_BracketingWithNeville, double ExpansionFactor_BracketingWithNeville, int NumVertices_BracketingWithNeville, int MAXIT_BracketingWithNeville)
        {
            if (TOL_Romberg <= 0 || NMIN_Romberg <= 0 || TOL_BracketingWithNeville <= 0 || ExpansionFactor_BracketingWithNeville <= 0 || NumVertices_BracketingWithNeville <= 0 || MAXIT_BracketingWithNeville <= 0)
                throw new System.Exception("Invalid input parameter(s) for NumUtils. All parameter must be greater than 0.");
            this.TOL_Romberg = TOL_Romberg;
            this.NMIN_Romberg = NMIN_Romberg;
            this.TOL_BracketingWithNeville = TOL_BracketingWithNeville;
            this.ExpansionFactor_BracketingWithNeville = ExpansionFactor_BracketingWithNeville;// to be used in the BracketingWithNeville routine.
            this.NumVertices_BracketingWithNeville = NumVertices_BracketingWithNeville;
            this.MAXIT_BracketingWithNeville = MAXIT_BracketingWithNeville;
            this.ROMBERG = new NumUtils.Integration.Romberg(this.NMIN_Romberg, this.TOL_Romberg);
            this.BWN = new NumUtils.RootFinding.BracketingWithNelville(this.ExpansionFactor_BracketingWithNeville, this.NumVertices_BracketingWithNeville, this.TOL_BracketingWithNeville, this.MAXIT_BracketingWithNeville);
            this.BWNR = new NumUtils.RootFinding.BracketingWithNewtonRaphson(this.ExpansionFactor_BracketingWithNeville, this.NumVertices_BracketingWithNeville, this.TOL_BracketingWithNeville, this.MAXIT_BracketingWithNeville);
        }

        public void ResetCosmoParameters(double OmegaM_0, double OmegaL_0, double OmegaR_0, double w_0, double h_0)
        {
            if (OmegaM_0 < 0 || OmegaL_0 < 0 || OmegaR_0 < 0 || h_0 < 0  )
                throw new System.Exception("Invalid input parameter(s) in CosmologicalFunctions. All of the following must apply: OmegaM_0 >= 0, OmegaL_0 >= 0, h_0 >=0.");
            this.OmegaM_0 = OmegaM_0;
            this.OmegaL_0 = OmegaL_0;
            this.OmegaR_0 = OmegaR_0;
            this.OmegaK_0 = Math.Round(1.0 - OmegaM_0 - OmegaL_0 - OmegaR_0, 14);
            this.OmegaT_0 = 1 - this.OmegaK_0;
            if (this.OmegaK_0 == 0)
                this.k = 0;
            else
                this.k = (double)(-Math.Sign(OmegaK_0));
            this.w_0 = w_0;
            this.h_0 = h_0;
            this.HubbleParameter_0 = this.h_0 * 100.0;
            this.HubbleDistance_0 = c / this.HubbleParameter_0;//in Mpc
            this.HubbleTime_0 = (c / this.HubbleParameter_0)*1e6 * 3.26163626/1e9;//in Gyr
            this.MaxTime = 0;
        }

        public void SetTolerances(double TOL_BracketingWithNeville, double TOL_Romberg, int NMIN_Romberg)
        {
            this.TOL_BracketingWithNeville = TOL_BracketingWithNeville;
            this.TOL_Romberg = TOL_Romberg;
            this.NMIN_Romberg = NMIN_Romberg;
        }

        public void SetDc(double z)
        {
            this.z = z;
            this.D_c = ComovDistLineOfSight(0, z);
        }

        public void SetDcDm(double z)
        {
            SetDc(z);
            this.D_m = ComovDistTransverseFromDc(this.D_c);
        }

        public void SetDcDmDaDlComVol(double z)
        {
            SetDcDm(z);
            this.D_a = AngularDiameterDist(this.z, this.D_m);
            this.D_l = LuminosityDistance(this.z, this.D_m);
            this.ComovingVolume = ComovVolume(this.z);//in Mpc^3
        }

        public void SetTimeIntervals(double z)
        {
            this.LOOKbackTIME = TimeInterval(0, z);
            this.AGEofUNIVERSE = TimeInterval(z, z_Max);
            this.TotalAGEofUNIVERSE = this.LOOKbackTIME + this.AGEofUNIVERSE;
        }

        public void SetAllCosmoQuantities(double z)
        {
            SetDcDmDaDlComVol(z);
            SetTimeIntervals(z);
        }

        public void SetPhi(double PhiStar, double MStar, double alpha)
        {
            this.PhiStar = PhiStar;
            this.MStar = MStar;
            this.Alpha = alpha;
        }

        public double E(double z)
        {
            return Math.Sqrt(OmegaM_0 * Math.Pow(1.0 + z, 3.0) + OmegaK_0 * Math.Pow(1.0 + z, 2.0) + OmegaL_0*Math.Pow( 1.0 + z,3*(1+StateEquationDarkEnergyEvolving(z))) +OmegaR_0*Math.Pow(1+z,4) );
        }

        public double E_squared(double z)
        {
            return (OmegaM_0 * Math.Pow(1.0 + z, 3.0) + OmegaK_0 * Math.Pow(1.0 + z, 2.0) + OmegaL_0 * Math.Pow(1.0 + z, 3 * (1 + StateEquationDarkEnergyEvolving(z))) + OmegaR_0 * Math.Pow(1 + z, 4));
        }

        public double E_inverse(double z)
        {
            return (1.0 / E(z));
        }

        public double E_inverse_derivative(double z)
        {
            return -0.5 * Math.Pow(OmegaM_0 * Math.Pow(1.0 + z, 3.0) + OmegaK_0 * Math.Pow(1.0 + z, 2.0) + OmegaL_0 * Math.Pow(1.0 + z, 3 * (1 + StateEquationDarkEnergyEvolving(z))) + OmegaR_0 * Math.Pow(1 + z, 4) , -1.5  ) *
                (3*OmegaM_0*(1+z)*(1+z)+2*OmegaR_0*(1+z)-3*(1+StateEquationDarkEnergyEvolving(z))*OmegaL_0*Math.Pow(1+z,-3*(1+StateEquationDarkEnergyEvolving(z))-1)        );
        }
        
        public double NormalizedScaleFactor(double Redshift)
        {
            return 1.0 / (1.0 + Redshift);
        }

        public double RecedingSpeedFromHubbleLaw(double z)
        {
            return c * z;
        }

        public double RecedingSpeedFromDoppler(double z)
        {
            double Z = (1+z)*(1+z);
            return c*( Z-1)/(Z+1);
        }

        public double RecedingSpeedFromUniverseExpansion(double z)
        {
            double Z = (1 + z) * (1 + z);
            return c * (Z - 1) / (Z + 1);
        }

        public double RedshiftAtDimensionlessScaleFactor(double NormalizedScaleFactor)
        {
            return 1.0 / NormalizedScaleFactor - 1.0;
        }

        public double HubbleParameter(double z)
        {// in Km/s/MPc
            return this.HubbleParameter_0 * E(z);
        }

        public double HubbleTime(double z)
        {
            return HubbleDistance(z) * 1e6 * 3.26163626 / 1e9;//in Gyr
        }

        public double HubbleTime_Derivative(double z)
        {
            return HubbleDistance_Derivative(z) * 1e6 * 3.26163626 / 1e9;//in Gyr
        }

        
        public double HubbleDistance(double z)
        {
            return c / HubbleParameter(z);//in Mpc
        }

        public double HubbleDistance_Derivative(double z)
        {
            return c / HubbleParameter_0 * E_inverse_derivative(z);//in Mpc
        }

        public double OmegaMatter(double z)
        {
            return ( this.OmegaM_0 *Math.Pow( (1+z),3)/ E_squared(z) );
        }

        public double OmegaRadiation(double z)
        {
            return (this.OmegaR_0 * Math.Pow((1 + z), 4) / E_squared(z));
        }

        public double OmegaLambda(double z)
        {
            return (this.OmegaL_0 * Math.Pow((1 + z), 3*(1+StateEquationDarkEnergyEvolving(z)) ) / E_squared(z));
        }

        public double OmegaTotal(double z)
        {
            //return OmegaMatter(z)+OmegaRadiation(z)+OmegaLambda(z);
            return (  this.OmegaM_0 *Math.Pow( (1+z),3)+  this.OmegaR_0 * Math.Pow((1 + z), 4) + this.OmegaL_0 * Math.Pow((1 + z), 3 * (1 + StateEquationDarkEnergyEvolving(z)))   )/ E_squared(z)    ;
        }

        public double OmegaCurvature(double z)
        {
            //return 1- OmegaTotal(z);
            return (this.OmegaK_0 * Math.Pow((1 + z), 2)) / E_squared(z);
        }

        public double RadiusOfUniverse(double z)
        {
            if (this.k == 0)
                return double.PositiveInfinity;
            else
                return HubbleDistance_0 * Math.Sqrt(this.k / -OmegaK_0) / (1 + z);
        }

        public double q0()
        {
            return this.OmegaR_0+this.OmegaM_0/2.0-this.OmegaL_0;// maybe to be redefined...
        }

        public double StateEquationDarkEnergyConstant(double z)
        {
            return this.w_0;
        }

        public double StateEquationDarkEnergyEvolving(double z)
        {
            return StateEquationDarkEnergyConstant(z);// to be changed in the future for the real evolving state eq.
        }

        double ParticleHorizon(double z)
        {
            if (z < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.ParticleHorizon. Redshift should be equal or greater than 0 and smaller than " + z_Max.ToString());
            return ComovDistLineOfSight(z);
        }

        public double LuminosityDistance(double z)
        {
            if (z < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.LuminosityDistance. Redshift should be equal or greater than 0 and smaller than " + z_Max.ToString());
            return (1 + z) * ComovDistTransverse(z);
        }

        public double LuminosityDistance(double z, double Dm)
        {
            if (z < 0 || Dm < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter(s) in CosmologicalFunctions.LuminosityDistance. Dm and z should be equal or greater than 0 and z smaller than " + z_Max.ToString());
            return  (1 + z) * Dm;
        }

        public double LuminosityDistance_Derivative(double z)
        {
            if (z < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.LuminosityDistance_Derivative. Redshift should be equal or greater than 0 and smaller than " + z_Max.ToString());
            return  (  (1 + z) * ComovDistTransverse_Derivative(z) + ComovDistTransverse(z)  );
        }

        public double ComovDistLineOfSight(double z)
        {
            if (z < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.ComovDistLineOfSight. Redshift should be equal or greater than 0 and smaller than " + z_Max.ToString());
            double Dc=0.0;
            if (z !=0 )
            {
                Dc = ComovDistLineOfSight(0, z);
            }
            return Dc;
        }

        public double ComovDistLineOfSight(double zMin, double zMax)
        {
            if (zMin < 0 || zMin > zMax)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.ComovDistLineOfSight. zMin should be equal or greater than 0 and equal or smaller than zMax.");
            double Dc = 0.0;
            double z1 = 0.0;
            double z2 = 0.0;
            if (zMax != 0 && zMin != zMax)
            {
                z1 = zMin;
                for (int i = 1; z2 <= zMax;i++)
                {
                    z2 = z1 + Math.Pow(DeltaZIntegration_Dc,i);// partitionig the redshift space into intervals.
                    if (z2 <= zMax)
                        Dc = Dc + (ROMBERG.RombergClosed(z1, z2, E_inverse));
                    else
                        Dc = Dc + (ROMBERG.RombergClosed(z1, zMax, E_inverse));
                    z1 = z2;
                }
            }
            Dc = Dc * this.HubbleDistance_0;
            return Dc;
        }

        public double ComovDistLineOfSight_Derivative(double z)
        {
            if (z < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.ComovDistLineOfSight_Derivative. Redshift should be equal or greater than 0 and smaller than " + z_Max.ToString());
            return this.HubbleDistance_0*E_inverse(z);
        }

        public double ComovDistTransverse(double z)
        {
            if (z < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.ComovDistTransverse. Redshift should be equal or greater than 0 and smaller than " + z_Max.ToString());
            double Dc = ComovDistLineOfSight(0,z);
            return ComovDistTransverseFromDc(Dc);
        }

        public double ComovDistTransverseFromDc(double Dc)
        {
            if (Dc < 0)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.ComovDistTransverseFromDc. Dc should be equal or greater than 0.");
            if (this.k == 0)
                return Dc;
            else
                return this.HubbleDistance_0 / Math.Sqrt(Math.Abs(this.OmegaK_0)) * S_k(Dc, this.OmegaK_0);
        }

        public double S_k(double r, int _k)
        {
            double Sk = 0;
            if (_k == 0)
                Sk = r;
            else
            {
                if (_k < 0)
                    Sk = Math.Sinh(r);
                else
                    Sk = Math.Sin(r);
            }            
            return Sk;
        }

        public double S_k(double CosmoDistance, double OMEGA_K)
        {
            double Sk=0;
            if (OMEGA_K == 0)
                Sk = CosmoDistance;
            else
            {
                if (OMEGA_K < 0)
                    Sk = Math.Sin(Math.Sqrt(Math.Abs(OMEGA_K)) * CosmoDistance / this.HubbleDistance_0);
                else
                    Sk = Math.Sinh(Math.Sqrt(OMEGA_K) * CosmoDistance / this.HubbleDistance_0);
            }
            return Sk;
        }

        public double ComovDistTransverse_Derivative(double z)
        {
            if (z < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.ComovDistTransverse_Derivative. Redshift should be equal or greater than 0.");
            double Dm_Derivative = 0;
            double SqrtOmegaK = Math.Sqrt(Math.Abs(OmegaK_0));
            if (OmegaK_0 == 0)
                Dm_Derivative = ComovDistLineOfSight_Derivative(z);
            else
            {
                if (OmegaK_0 > 0)
                    Dm_Derivative = Math.Cosh(SqrtOmegaK * ComovDistLineOfSight(z) / HubbleDistance_0) * ComovDistLineOfSight_Derivative(z);    
                else
                    Dm_Derivative = Math.Cos(SqrtOmegaK * ComovDistLineOfSight(z) / HubbleDistance_0) * ComovDistLineOfSight_Derivative(z); //OmegaK_0<0
            }
            return Dm_Derivative;
        }

        public double ComovDistBetween2objectsThetaPhi(double Redshift1, double Theta1, double Phi1, double Redshift2, double Theta2, double Phi2)
        {
            double SeparationAngle = MATH.AngularSeparationThetaPhi(Theta1, Phi1, Theta2, Phi2);
            return ComovDistBetween2objects(Redshift1, Redshift2, SeparationAngle);
        }

        public double ComovDistBetween2objectsRaDec(double Redshift1, double RA1, double DEC1, double Redshift2, double RA2, double DEC2)
        {
            double SeparationAngle = MATH.AngularSeparationRaDec(RA1, DEC1, RA2, DEC2);
            return ComovDistBetween2objects(Redshift1, Redshift2, SeparationAngle);
        }

        public double ComovDistBetween2objectsXyz(double Redshift1, double x1, double y1, double z1, double Redshift2, double x2, double y2, double z2)
        {
            double SeparationAngle = MATH.AngularSeparationXyz(x1, y1, z1, x2, y2, z2);
            return ComovDistBetween2objects(Redshift1, Redshift2, SeparationAngle);
        }

        public double ComovDistBetween2objects2(double Redshift1, double Redshift2, double SeparationAngle)
        {
            double answer = 0;
            double s1 = ComovDistLineOfSight(Redshift1);// / this.HubbleDistance_0;
            double s2 = ComovDistLineOfSight(Redshift2);// / this.HubbleDistance_0;
            //double d = Math.Sqrt(Math.Pow(s2 * Math.Sin(SeparationAngle), 2) + Math.Pow(s2 * Math.Cos(SeparationAngle) * Math.Sqrt(1 + Math.Sign(OmegaK_0) * s1 * s1) - s1 * Math.Sqrt(1 + Math.Sign(OmegaK_0) * s2 * s2), 2));
            double d = Math.Sqrt(Math.Pow(s2 * Math.Sin(SeparationAngle), 2) + Math.Pow(s2 * Math.Cos(SeparationAngle) * Math.Sqrt(1 - this.k * s1 * s1) - s1 * Math.Sqrt(1 -this.k* s2 * s2), 2));

            if (this.OmegaK_0 == 0)
                answer = d;
            else
                if (OmegaK_0 > 0)
                    answer = Math.Sqrt(Math.Abs(OmegaK_0)) * Math.Sinh(d);
                else
                    answer = Math.Sqrt(Math.Abs(OmegaK_0)) * Math.Sin(d);
            return answer;//* this.HubbleDistance_0;
        }

        public double ComovDistBetween2objects(double Redshift1, double Redshift2, double SeparationAngle)
        {
            double answer = 0;
            double Dc1 = ComovDistLineOfSight(Redshift1);// / this.HubbleDistance_0;
            double Dc2 = ComovDistLineOfSight(Redshift2);// / this.HubbleDistance_0;

            if (this.k == 0)
                answer = Math.Sqrt(Dc1 * Dc1 + Dc2 * Dc2 - 2 * Dc1 * Dc2 * Math.Cos(SeparationAngle));
            else
            {
                double R = RadiusOfUniverse(0);
                Dc1 = Dc1 / R;
                Dc2 = Dc2 / R;
                if (this.k > 0)
                {
                    answer = R * Math.Acos(Math.Cos(Dc1) * Math.Cos(Dc2) + Math.Sin(Dc1) * Math.Sin(Dc2) * Math.Cos(SeparationAngle));
                }
                else
                {
                    answer = R * MATH.ACosh(Math.Cosh(Dc1) * Math.Cosh(Dc2) - Math.Sinh(Dc1) * Math.Sinh(Dc2) * Math.Cos(SeparationAngle));
                }
                //Console.WriteLine("radius2 = {0} {1} {2} {3}", R, Dc1, Dc2, answer);
            }
            return answer;// *this.HubbleDistance_0;
        }

        public double AngularDiameterDist(double z)
        {
            if (z < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.AngularDiameterDist. Redshift should be equal or greater than 0 and smaller than " + z_Max.ToString());
            return (ComovDistTransverse(z) / (1 + z));
        }
        
        public double AngularDiameterDist(double z, double Dm)
        {
            if (z < 0 || Dm<0)
                throw new System.Exception("Invalid paramter(s) in CosmologicalFunctions.AngularDiameterDist. Dm and z should be equal or greater than 0.");
            return Dm / (1 + z);
        }

        public double AngularDiameterDist_Derivative(double z)
        {
            if (z < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.AngularDiameterDist_Derivative. Redshift should be equal or greater than 0 and smaller than " + z_Max.ToString());
            return  ComovDistTransverse_Derivative(z)/(1+z) - ComovDistTransverse(z)/(1 + z)/(1+z);
        }

        public double ComovVolume(double z)
        {
            // prefered method to calculte the comoving volume:
            if (z < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.ComovVolume. Redshift should be equal or greater than 0 and smaller than " + z_Max.ToString());
            double Dc = ComovDistLineOfSight(z);
            return ComovVolumeFromDc(Dc);
        }

        public double ComovVolume_Derivative(double z)
        {
            // prefered method to calculte the comoving volume:
            if (z < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.ComovVolume_Derivative. Redshift should be equal or greater than 0 and smaller than " + z_Max.ToString());
            return 4.0 * Math.PI * this.HubbleDistance_0 * Math.Pow(ComovDistTransverse(z), 2.0) / E(z);
        }
        
        public double ComovVolumeFromDm(double Dm)
        {
            if (Dm < 0)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.ComovVolumeFromDm. Dm should be equal or greater than 0.");
            double ComVol = 0.0;
            if (OmegaK_0 == 0)
                ComVol = (4.0 / 3.0 * Math.PI * Math.Pow(Dm, 3));
            else
            {
                double DmDh = Dm / this.HubbleDistance_0;
                double arg1 = (4 * Math.PI * Math.Pow(this.HubbleDistance_0, 3)) / (2 * OmegaK_0);
                double arg2 = Math.Sqrt(1 + OmegaK_0 * Math.Pow(DmDh, 2));
                double arg3 = Math.Sqrt(Math.Abs(OmegaK_0)) * DmDh;
                if (OmegaK_0 > 0)
                    ComVol = arg1 * (DmDh * arg2 - Math.Pow(Math.Abs(OmegaK_0), -0.5) * MATH.ASinh(arg3));
                else
                    ComVol = arg1 * (DmDh * arg2 - Math.Pow(Math.Abs(OmegaK_0), -0.5) * Math.Asin(arg3));
            }
            return ComVol;//in Mpc^3
        }

        public double ComovVolumeFromDc(double Dc)
        {
            if (Dc < 0)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.ComovVolumeFromDc. Dc should be equal or greater than 0.");
            double ComVol = 0.0;
            if (OmegaK_0 == 0)
                ComVol = (4.0 / 3.0 * Math.PI * Math.Pow(Dc, 3));
            else
            {
                double R_0 = RadiusOfUniverse(0);
                double r = Dc / R_0;
                if (this.OmegaK_0 > 0)
                    ComVol = 2 * Math.PI * R_0 * R_0 * R_0 * (r - Math.Sin(r) * Math.Cos(r));
                else
                    ComVol = 2 * Math.PI * R_0 * R_0 * R_0 * (Math.Sinh(r) * Math.Cosh(r) - r);
            }
            return ComVol;//in Mpc^3
        }
        
        public double TimeInterval(double zMin, double zMax)
        {
            if (zMin >= 0 && zMax >= zMin && zMax <= z_Max)
            {
                double time;
                if (zMin >= zMiddle)
                {
                    time = (ROMBERG.RombergOpenInfinity(zMin, zMax, new NumUtils.DelegateFunction.DoubleFun(IntegrandTimeInterval)));// in years
                }
                else
                {
                    if (zMax <= zMiddle)
                    {
                        time = (ROMBERG.RombergClosed(zMin, zMax, new NumUtils.DelegateFunction.DoubleFun(IntegrandTimeInterval)));// in years
                    }
                    else
                    {
                        time = (ROMBERG.RombergClosed(zMin, zMiddle, new NumUtils.DelegateFunction.DoubleFun(IntegrandTimeInterval)));
                        time = time + (ROMBERG.RombergOpenInfinity(zMiddle, zMax, new NumUtils.DelegateFunction.DoubleFun(IntegrandTimeInterval)));
                    }
                }
                time = time * this.HubbleTime_0;//in Gyr
                return time;
            }
            else
                throw new System.Exception("Invalid paramter(s) in CosmologicalFunctions.TimeInterval. zMin should be equal or greater than 0 and zMax should be equal or greater than zMin and smaller than " + z_Max.ToString());
        }

        public double LookBackTime(double z)
        {
            return TimeInterval(0, z);
        }

        public double LookBackTime_Derivative(double z)
        {
            if (z < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.LookBackTime_Derivative. Redshift should be equal or greater than 0 and smaller than " + z_Max.ToString());
            return this.HubbleTime_0*IntegrandTimeInterval(z);
        }

        public double AgeOfUniverse(double z)
        {
            return TimeInterval(z, z_Max);
        }

        public double AgeOfUniverse_Derivative(double z)
        {
            if (z < 0 || z > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.AgeOfUniverse_Derivative. Redshift should be equal or greater than 0 and smaller than " + z_Max.ToString());
            return -LookBackTime_Derivative(z);
        }

        private double IntegrandTimeInterval(double z)
        {
            return (1 / ((1 + z) * E(z)));
        }

        public double AbsMagFromLumDist(double m, double LuminosityDistance, double Kcorr)
        {
            if (LuminosityDistance <= 0)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.AbsMagFromLumDist. Luminosity Distance should be greater than 0.");
            // D_l in Mpc, as given in  <double D_l(...)>
            return (m - 5 * Math.Log10(LuminosityDistance) - 25 - Kcorr);
        }

        public double AbsMag(double m, double z, double Kcorr)
        {
            if (z <= 0 || z > z_Max)
                throw new System.Exception("Invalid parameter in CosmologicalFunctions.AbsMag. Redshift should be greater than 0 and smaller than " + z_Max.ToString());
            return (m - 5 * Math.Log10(LuminosityDistance(z)) - 25 - Kcorr);
        }

        public double RedshiftAtHubbleTime(double HubbleTIME)
        {
            if (HubbleTIME < 0)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.RedshiftAtHubbleTime. HubbleTIME should be equal or greater than 0.");
            return BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(HubbleTime), new NumUtils.DelegateFunction.DoubleFun(HubbleTime_Derivative), HubbleTIME, Redshift0_Bracketing, Redshift1_Bracketing);
        }

        public double RedshiftAtHubbleDistance(double HubbleDISTANCE)
        {
            if (HubbleDISTANCE < 0)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.RedshiftAtHubbleDistance. HubbleDISTANCE should be equal or greater than 0.");
            return BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(HubbleDistance), new NumUtils.DelegateFunction.DoubleFun(HubbleDistance_Derivative), HubbleDISTANCE, Redshift0_Bracketing, Redshift1_Bracketing);
        }

        public double RedshiftAtComovDistLineOfSight(double ComovingDistLineOfSight)
        {
            if (ComovingDistLineOfSight < 0)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.RedshiftAtComovDistLineOfSight. ComovDistLineOfSight should be equal or greater than 0.");
            //return BWN.BracketingWithNelvilleRootFinding(new NumUtils.DelegateFunction.DoubleFun(ComovDistLineOfSight), ComovingDistLineOfSight, Redshift0_Bracketing, Redshift1_Bracketing);
            return BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(ComovDistLineOfSight), new NumUtils.DelegateFunction.DoubleFun(ComovDistLineOfSight_Derivative), ComovingDistLineOfSight, Redshift0_Bracketing, Redshift1_Bracketing);
        }

        public double RedshiftAtLuminosityDistance(double LumDist)
        {
            if (LumDist < 0)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.RedshiftAtLuminosityDistance. LumDist should be equal or greater than 0.");
            //return BWN.BracketingWithNelvilleRootFinding(new NumUtils.DelegateFunction.DoubleFun(LuminosityDistance), LumDist, Redshift0_Bracketing, Redshift1_Bracketing);
            return BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(LuminosityDistance), new NumUtils.DelegateFunction.DoubleFun(LuminosityDistance_Derivative), LumDist, Redshift0_Bracketing, Redshift1_Bracketing);
        }

        public double RedshiftAtComovDistTransverse(double ComovDistTrans)
        {
            if (ComovDistTrans < 0)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.RedshiftAtComovDistTransverse. ComovDistTrans should be equal or greater than 0.");
            //return BWN.BracketingWithNelvilleRootFinding_ForSmoothFunction(new NumUtils.DelegateFunction.DoubleFun(ComovDistTransverse), ComovDistTrans, Redshift0_Bracketing, Redshift1_Bracketing);
            return BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(ComovDistTransverse), new NumUtils.DelegateFunction.DoubleFun(ComovDistTransverse_Derivative), ComovDistTrans, Redshift0_Bracketing, Redshift1_Bracketing);
        }

        public double RedshiftAtMaxOfAngularDiamDist()
        {
            return BWN.BracketingWithNelvilleRootFinding(new NumUtils.DelegateFunction.DoubleFun(AngularDiameterDist_Derivative), 0.0, Redshift0_Bracketing, Redshift1_Bracketing);
        }

        public double[] RedshiftAtAngularDiamDist(double AngularDiamDist)
        {
            if (AngularDiamDist < 0)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.RedshiftAtAngularDiamDist. AngularDiamDist should be equal or greater than 0.");
            if (AngularDiamDist == 0)
                return new double[2] { 0, 1.7976931348623e308};
            double[] Answer = new double[2];
            this.zAtDaMax = RedshiftAtMaxOfAngularDiamDist();
            this.DaMax = AngularDiameterDist(zAtDaMax);
            //Answer[0] = BWN.BracketingWithNelvilleRootFinding(new NumUtils.DelegateFunction.DoubleFun(AngularDiameterDist_zAtDaMax_1), AngularDiamDist, 0.0, this.zAtDaMax);
            //Answer[1] = BWN.BracketingWithNelvilleRootFinding(new NumUtils.DelegateFunction.DoubleFun(AngularDiameterDist_zAtDaMax_2), AngularDiamDist, this.zAtDaMax, this.zAtDaMax * 2);
            Answer[0] = BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(AngularDiameterDist_zAtDaMax_1), new NumUtils.DelegateFunction.DoubleFun(AngularDiameterDist_Derivative), AngularDiamDist, 0.0, this.zAtDaMax);
            Answer[1] = BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(AngularDiameterDist_zAtDaMax_2), new NumUtils.DelegateFunction.DoubleFun(AngularDiameterDist_Derivative), AngularDiamDist, this.zAtDaMax, this.zAtDaMax * 2);
            return Answer;
        }

        private double AngularDiameterDist_zAtDaMax_1(double z)
        {
            if (z > this.zAtDaMax) { return this.DaMax + Math.Exp(z - this.zAtDaMax); }// Math.Pow((z - this.zAtDaMax), 2.0); }
            else { return AngularDiameterDist(z); }
        }

        private double AngularDiameterDist_zAtDaMax_2(double z)
        {
            if (z < this.zAtDaMax) { return this.DaMax + Math.Exp(this.zAtDaMax - z); }// Math.Pow((z - this.zAtDaMax), 2.0); }
            else { return AngularDiameterDist(z); }
        }

        public double RedshiftAtAgeOfUniverse(double AGEofUniverse)
        {
            this.MaxTime = TimeInterval(0, z_Max);
            if (AGEofUniverse < 0 || AGEofUniverse > this.MaxTime)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.RedshiftAtAgeOfUniverse. AgeOfUniverse should exist only inside the interval (0," + MaxTime.ToString() + "].");
            if (AGEofUniverse == this.MaxTime) { return 0.0; }
            if (AGEofUniverse ==0 || 1 / AGEofUniverse >= z_Max) { return 1.7976931348623e308; }
            //return BWN.BracketingWithNelvilleRootFinding(new NumUtils.DelegateFunction.DoubleFun(AgeOfUniverse), AGEofUniverse, Redshift0_Bracketing, Redshift1_Bracketing);
            //return BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(AgeOfUniverse),new NumUtils.DelegateFunction.DoubleFun(AgeOfUniverse_Derivative), AGEofUniverse, Redshift0_Bracketing, Redshift1_Bracketing);
            if (AGEofUNIVERSE > this.MaxTime / 2)
            {
                this.ChooseTimeFunction = 0;
                return BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(Time), new NumUtils.DelegateFunction.DoubleFun(AgeOfUniverse_Derivative), AGEofUniverse, Redshift0_Bracketing, Redshift1_Bracketing);
            }
            else
                return BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(AgeOfUniverse), new NumUtils.DelegateFunction.DoubleFun(AgeOfUniverse_Derivative), AGEofUniverse, Redshift0_Bracketing, Redshift1_Bracketing);
        }

        public double RedshiftAtLookBackTime(double LookBackTIME)
        {
            //if (this.MaxTime == 0)
                this.MaxTime = TimeInterval(0, z_Max);
            if (LookBackTIME < 0 || LookBackTIME > this.MaxTime)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions.RedshiftAtLookBackTime. LookBackTime should exist only inside the interval [0," + MaxTime.ToString() + "].");
            if (LookBackTIME == this.MaxTime) { return 1.7976931348623e308; }
            if (LookBackTIME == 0 || 1 / LookBackTIME >= z_Max) { return 0; }
            //return BWN.BracketingWithNelvilleRootFinding(new NumUtils.DelegateFunction.DoubleFun(LookBackTime), LookBackTIME, Redshift0_Bracketing, Redshift1_Bracketing);
            if (LookBackTIME < this.MaxTime / 2)
                return BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(LookBackTime), new NumUtils.DelegateFunction.DoubleFun(LookBackTime_Derivative), LookBackTIME, Redshift0_Bracketing, Redshift1_Bracketing);
            else
            {
                this.ChooseTimeFunction = 1;
                return BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(LookBackTime), new NumUtils.DelegateFunction.DoubleFun(LookBackTime_Derivative), LookBackTIME, Redshift0_Bracketing, Redshift1_Bracketing);
                //return BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(Time), new NumUtils.DelegateFunction.DoubleFun(AgeOfUniverse_Derivative), LookBackTIME, Redshift0_Bracketing, Redshift1_Bracketing);
            }
        }

        private double Time(double z)
        {
            if (ChooseTimeFunction == 0)
                return this.MaxTime - LookBackTime(z);
            else
                return this.MaxTime - AgeOfUniverse(z);
        }

        public double RedshiftAtComovVolume(double ComovVOLUME)
        {
            if (ComovVOLUME < 0 || ComovVOLUME > z_Max)
                throw new System.Exception("Invalid paramter in CosmologicalFunctions. ComovVolume should be greater than 0");
            //return BWN.BracketingWithNelvilleRootFinding(new NumUtils.DelegateFunction.DoubleFun(ComovVolume), ComovVOLUME, Redshift0_Bracketing, Redshift1_Bracketing);
            return BWNR.GetRootWithNewtonRaphsonAndNeville(new NumUtils.DelegateFunction.DoubleFun(ComovVolume), new NumUtils.DelegateFunction.DoubleFun(ComovVolume_Derivative), ComovVOLUME, Redshift0_Bracketing, Redshift1_Bracketing);
        }

        public double ComovVolumeAtLumDist(double LuminosityDistance)
        {
            if (LuminosityDistance < 0)
                throw new System.Exception("Invalid paramter(s) in CosmologicalFunctions.ComovVolumeAtLumDist. LuminosityDistance must be greater or equal than 0.");
            double z_answer = RedshiftAtLuminosityDistance(LuminosityDistance);
            return ComovVolume(z_answer);
        }

        public IEnumerable GridAllCosmoQuantities(double zMin, double zMax, int NumCell)
        {

            if (zMin < 0 || zMin > zMax || NumCell < 0 || zMax > z_Max)
                throw new System.Exception("Invalid paramter(s) in CosmologicalFunctions.GridAllCosmoQuantitiesIEnumerable. Usage: zMin>=0, zMin<=zMax and NumCell>=0 and zMax<" + z_Max.ToString());

            double[] CosmoQuantities = new double[8];
            double TotalAgeOfUniverse = TimeInterval(0, z_Max);
            double StoredZ,StoredDc,StoredLBT;

            double DeltaZ = (zMax - zMin) / NumCell;
            double Redshift = zMin;

            CosmoQuantities[0] = Redshift;//z
            CosmoQuantities[1] = ComovDistLineOfSight(0, Redshift);//Dc
            CosmoQuantities[2] = ComovDistTransverseFromDc(CosmoQuantities[1]);//Dm
            CosmoQuantities[3] = AngularDiameterDist(Redshift, CosmoQuantities[2]);//Da
            CosmoQuantities[4] = LuminosityDistance(Redshift, CosmoQuantities[2]);//Dl
            CosmoQuantities[5] = ComovVolumeFromDm(CosmoQuantities[2]);//ComovingVolume
            CosmoQuantities[6] = TimeInterval(0, Redshift);// LookBackTime
            CosmoQuantities[7] = TotalAgeOfUniverse - CosmoQuantities[6];//AgeOfUniverse
            StoredZ = CosmoQuantities[0]; StoredDc = CosmoQuantities[1]; StoredLBT = CosmoQuantities[6];
            yield return CosmoQuantities;

            for (int i = 1; i < (NumCell + 1); i++)
            {
                Redshift = MATH.Round(zMin + i * DeltaZ, SignificantPlaces); 
                CosmoQuantities[0] = Redshift;
                CosmoQuantities[1] = StoredDc + ComovDistLineOfSight(StoredZ, Redshift);//Dc
                CosmoQuantities[2] = ComovDistTransverseFromDc(CosmoQuantities[1]);//Dm
                CosmoQuantities[3] = AngularDiameterDist(Redshift, CosmoQuantities[2]);//Da
                CosmoQuantities[4] = LuminosityDistance(Redshift, CosmoQuantities[2]);//Dl
                CosmoQuantities[5] = ComovVolumeFromDm(CosmoQuantities[2]);//ComovingVolume
                CosmoQuantities[6] = StoredLBT + TimeInterval(StoredZ, Redshift);// LookBackTime
                CosmoQuantities[7] = TotalAgeOfUniverse - CosmoQuantities[6];//AgeOfUniverse
                StoredZ = CosmoQuantities[0]; StoredDc = CosmoQuantities[1]; StoredLBT = CosmoQuantities[6];
                yield return CosmoQuantities;
            }
        }

        public double SchechterFunction(double M)
        {
            double dif = (0.4 * (MStar - M));
            double arg = (-1 * Math.Pow(10, dif));
            double al = Alpha + 1;
            double a = (0.4 * 2.30258509299405 * PhiStar * Math.Pow(arg, Alpha + 1) * Math.Exp(arg));
            return a;
        }

        public double SchechterFunction(double M, double PhiStar, double Mstar, double alpha)
        {
            double dif = 0.4 * (Mstar - M);
            double arg = (Math.Pow(10, dif));
            double a2 = (alpha + 1.0);
            double a = (0.4 * 2.30258509299405 * PhiStar * Math.Pow(arg, a2) * Math.Exp(-arg));
            return a;
        }

        public double SchechterFunction(double M, double Mstar, double alpha)
        {
            double dif = 0.4 * (Mstar - M);
            double arg = (Math.Pow(10, dif));
            double a2 = (alpha + 1.0);
            double a = (0.4 * 2.30258509299405 * Math.Pow(arg, a2) * Math.Exp(-arg));
            return a;
        }

        static public string Revision
        {
            get { return "$Revision: 1.0 $"; }
        }

        /* Revision History

                $Log: CosmologicalFunctions.cs,v $
                Revision 1.1  2008/06/20 10:59:55  mtaghiza
                 Created Revision and versions for each class.
        */

        public string[] Revisions()
        {
            string[] revs = {
					"CosmologicalFunctions "+ CosmologicalFunctions.Revision, 
					"HaloFunctions "+ HaloOccupationFunctions.Revision,
                    "SQLCosmologicalFunctionsFixedParameters " + SQLCosmology.SQLCosmologicalFunctionsFixedParameters.Revision,
                    "SQLCosmologicalFunctionsFreeParameters " + SQLCosmology.SQLCosmologicalFunctionsFreeParameters.Revision,
					};
            return revs;
        }


    }




}


