using System;
using System.Collections.Generic;
using System.Data.SqlClient;
using System.Collections;

namespace Mathematics
{

    public class MathFunctions
    {
        private int digits = 14;

        public double AngularSeparationRaDec(double RA1, double DEC1, double RA2, double DEC2)
        {
            return AngularSeparationThetaPhi((90 - DEC1) * Math.PI / 180, RA1 * Math.PI / 180, (90 - DEC2) * Math.PI / 180, RA2 * Math.PI / 180);
        }

        public double AngularSeparationXyz(double x1, double y1, double z1, double x2, double y2, double z2)
        {
            double DotProduct = x1 * x2 + y1 * y2 + z1 * z2;
            double N1 = x1 * x1 + y1 * y1 + z1 * z1;
            double N2 = x2 * x2 + y2 * y2 + z2 * z2;
            DotProduct = DotProduct / Math.Sqrt(N1 * N2);
            if (DotProduct > 1)
                DotProduct = 1;
            if (DotProduct < -1)
                DotProduct = -1;
            return Math.Acos(DotProduct);
        }

        public void ThetaPhi2Xyz(double Theta, double Phi, out double x, out double y, out double z)
        {
            double sint = Math.Sin(Theta);
            x = sint * Math.Cos(Phi);
            y = sint * Math.Sin(Phi);
            z = Math.Cos(Theta);
        }

        public double AngularSeparationThetaPhi(double Theta1, double Phi1, double Theta2, double Phi2)
        {
            double x1, x2, y1, y2, z1, z2;
            ThetaPhi2Xyz(Theta1, Phi1, out x1, out y1, out z1);
            ThetaPhi2Xyz(Theta2, Phi2, out x2, out y2, out z2);
            return AngularSeparationXyz(x1, y1, z1, x2, y2, z2);
        }

        public long BitShiftRight(long LongInteger, int shift)
        {
            return (LongInteger >> shift);
        }

        public long BitShiftLeft(long LongInteger, int shift)
        {
            return (LongInteger << shift);
        }

        public double Random(int seed)
        {
            Random RandomClass = new Random(seed);
            return RandomClass.NextDouble();
        }

        public double ASinh(double x)
        { 
            //note: arcsinh(x)= Ln(x + Sqrt(x*x + 1) ), not defined in the class "Math"
            return Math.Log(x + Math.Sqrt(x * x + 1.0));
        }

        public double ACosh(double x)
        {
            //note: arccosh(x)= Ln(x + Sqrt(x*x - 1) ), not defined in the class "Math"
            if (x < 1)
                throw new System.Exception("Invalid parameter in ACosh. x>=1 must apply");
            else
                return Math.Log(x + Math.Sqrt(x * x - 1.0));
        }

        public double[] EulerRotation(double[] X, double phi, double theta, double psi)
        {
            double[] RotatedX = new double[3];
            double[,] A = new double[3, 3];
            A[0, 0] = Math.Cos(psi) * Math.Cos(phi) - Math.Cos(theta) * Math.Sin(phi) * Math.Sin(psi);
            A[0, 1] = Math.Cos(psi) * Math.Sin(phi) + Math.Cos(theta) * Math.Cos(phi) * Math.Sin(psi);
            A[0, 2] = Math.Sin(psi) * Math.Sin(theta);
            A[1, 0] = -Math.Sin(psi) * Math.Cos(phi) - Math.Cos(theta) * Math.Sin(phi) * Math.Cos(psi);
            A[1, 1] = -Math.Sin(psi) * Math.Sin(phi) + Math.Cos(theta) * Math.Cos(phi) * Math.Cos(psi);
            A[1, 2] = Math.Cos(psi) * Math.Sin(theta);
            A[2, 0] = Math.Sin(theta) * Math.Sin(phi);
            A[2, 1] = -Math.Sin(theta) * Math.Cos(phi);
            A[2, 2] = Math.Cos(theta);

            for (int i = 0; i < 3; i++)
            {
                for (int j = 0; j < 3; j++)
                {
                    RotatedX[i] = RotatedX[i] + A[i, j] * X[j];
                }
            }
            return RotatedX;
        }

        public double Round(double x, int digits)
        {
            if (x == 0 || digits <= 0)
                return 0;
            if(digits >= 17)
                throw new System.Exception("the value of 'digits' has to be smaller or equal than 16");
            else
            {
                int z = 0;
                if (Math.Abs(x) > 1)
                {
                    z = (int)Math.Pow(10, (int)Math.Log10(Math.Abs(x)));
                    return z * Math.Round(x / z, digits - 1);
                }
                else
                {
                    z = (int)Math.Pow(10, 1 - (int)Math.Log10(Math.Abs(x)));
                    return Math.Round(x * z, digits - 1) / z;
                }
            }
        }

        public IEnumerable Grid(double X1, double X2, int NumCell, bool IsLinear, bool IsMidPoints)
        {
            if (X1 >= X2 || NumCell <= 0)
                throw new System.Exception("Invalid paramter(s) in MathFunctions.Grid. X1 should be less than X2, while NumCell should be greater than zero.");
            if ( X1 <=0 && IsLinear == false)
                throw new System.Exception("Invalid paramter(s) in MathFunctions.Grid. If IsLinear=false (i.e, log10 logaritmic scale), then X1 should be greater than zero.");

            double[] Grid = new double[1];
            double DeltaX;
            if (IsLinear == true)
            {
                DeltaX = (X2 - X1) / NumCell;
                if (IsMidPoints)
                {
                    for (int i = 0; i < (NumCell); i++)
                    {
                        Grid[0] = Round(X1 + (i + 0.5) * DeltaX, digits);
                        yield return Grid;
                    }
                }
                else
                {
                    for (int i = 0; i < (NumCell); i++)
                    {
                        Grid[0] = Round(X1 + i * DeltaX, digits);
                        yield return Grid;
                    }
                    Grid[0] = X2;
                    yield return Grid;
                }
            }
            else // log10 scale
            {
                
                if (IsMidPoints)
                {
                    double Pow10DeltaX = Math.Pow(10, (Math.Log10(X2) - Math.Log10(X1)) / NumCell);
                    double StoredX = X1 * Math.Sqrt(Pow10DeltaX);
                    Grid[0] = StoredX;
                    yield return Grid;
                    for (int i = 0; i < (NumCell-1); i++)
                    {
                        StoredX = StoredX * Pow10DeltaX;
                        Grid[0] = Round(StoredX, digits);
                        yield return Grid;
                    }
                }
                else
                {
                    double StoredX = X1;
                    DeltaX = Math.Pow((X2 / X1), 1.0 / (NumCell));
                    Grid[0] = X1;
                    yield return Grid;
                    for (int i = 1; i < (NumCell); i++)
                    {
                        StoredX = StoredX * DeltaX;
                        Grid[0] = Round(StoredX, digits);
                        yield return Grid;
                    }
                    Grid[0] = X2;
                    yield return Grid;
                }
            }
        }

        public double Bin(double X, double X1, double X2, int NumCell, bool HasOpenUpperBound, bool IsLinearScale)
        {
            if (X1 >= X2 || NumCell <= 0)
                throw new System.Exception("Invalid paramter(s) in MathFunctions.GetBin. X1 should be less than X2 and NumCell greater than zero");
            if(!IsLinearScale && (X<=0 || X1<=0))
                throw new System.Exception("Invalid paramter(s) in MathFunctions.GetBin. When logarithmic (base 10) scale is chosen, all X,X1 and X2 should be greater than 0");
            
            if (IsLinearScale)
            {
                double DeltaX = ((X2 - X1) / NumCell);
                int ind = 0;
                if (HasOpenUpperBound == true)
                {
                    ind = (int)Math.Floor(Round((X - X1) / DeltaX, digits));
                }
                else
                {
                    ind = (int)Math.Ceiling(Round((X - X1) / DeltaX, digits)) - 1;
                }
                return Round(X1 + (DeltaX * (ind + 0.5)), digits);
            }
            else // in Log10 scale
            {
                double LogX1 = Math.Log10(X1);
                double LogX2 = Math.Log10(X2);
                double DeltaX = (LogX2 - LogX1) / NumCell;
                int ind = 0;
                if (HasOpenUpperBound == true)
                {
                    ind = (int)Math.Floor(Round((Math.Log10(X) - LogX1) / DeltaX, digits));
                }
                else
                {
                    ind = (int)Math.Ceiling(Round((Math.Log10(X) - LogX1) / DeltaX, digits)) - 1;
                }
                return Round(Math.Pow(10, X1 + DeltaX * (ind - 0.5)), digits);
            }
        }

        public double[,] Histogram(ref ArrayList X, ref ArrayList Weights, double X1, double X2, int NumCell, bool HasOpenUpperBound)
        {
            // implements histogram with closed lower bound and open upper bound or the reverse

            if (X1 >= X2 || NumCell <= 0)
                throw new System.Exception("Invalid paramter(s) in MathFunctions.Histogram. X1 should be less than X2 and NumCell greater than zero");

            double[,] Histogram = new double[NumCell, 3];
            int ind = 0;
            double DeltaX = ((X2 - X1) / NumCell);
            DeltaX = Round(DeltaX, digits);
            double area = 0;
            double X_value = 0;
            if (HasOpenUpperBound == true)
            {
                for (int i = 0; i < X.Count; i++)
                {
                    X_value = (double)X[i];
                    if (X_value >= X1 && X_value < X2)
                    {
                        //ind = (int)Math.Floor(NumCell * ((double)X[i] - X1) / IntervalX);
                        ind = (int)Math.Floor(Round((X_value - X1) / DeltaX, digits));
                        Histogram[ind, 1] = Histogram[ind, 1] + (double)Weights[i];
                        Histogram[ind, 2] = Histogram[ind, 2] + 1;
                    }
                }
            }
            else
            {
                for (int i = 0; i < X.Count; i++)
                {
                    X_value = (double)X[i];
                    if (X_value > X1 && X_value <= X2)
                    {

                        ind = (int)Math.Ceiling(Round((X_value - X1) / DeltaX, digits)) - 1;
                        Histogram[ind, 1] = Histogram[ind, 1] + (double)Weights[i];
                        Histogram[ind, 2] = Histogram[ind, 2] + 1;
                    }
                }
            }
            //filling up the X cells 
            for (int i = 0; i < NumCell; i++)
            {
                Histogram[i, 0] = X1 + (double)(DeltaX * ((double)(i) + 0.5));
                area = area + Histogram[i, 1] * DeltaX;
            }
            //Normalizing:
            for (int i = 0; i < NumCell; i++)
            {
                Histogram[i, 1] = Histogram[i, 1] / area;
            }

            return Histogram;

        }

        public double[,] HistogramNDimFromQuery(string SQLConnection, string QueryToTable, int Dimension, string String_X1, string String_X2, string String_NumCell, bool HasOpenUpperBound)
        {
            ArrayList Data = new ArrayList(Dimension);
            for (int i = 0; i < Dimension; i++)
            {
                Data.Add(new ArrayList());
            }
            ArrayList Weights = new ArrayList();

            using (SqlConnection connection = new SqlConnection(SQLConnection))
            using (SqlCommand cmd = new SqlCommand(QueryToTable, connection))
            {
                try
                {
                    connection.Open();
                    SqlDataReader rdr = cmd.ExecuteReader();
                    //SqlContext.Pipe.Send("defergerger 1");
                    rdr.Read();
                    //SqlContext.Pipe.Send("defergerger 1.5 " + rdr.FieldCount.ToString());
                    if (rdr.FieldCount > (Dimension + 1) || rdr.FieldCount == 0)
                        throw new System.Exception("Incorrect number of columns in SQLMathCLR.SQLMathFunctions.HistogramNDimFromQuery. Number of Columns returned by QueryToTable should be between 1 and (Dimesion+1).");
                    if (rdr.FieldCount == Dimension)
                    {
                        //SqlContext.Pipe.Send("defergerger 2");
                        for (int i = 0; i < Dimension; i++) { ((ArrayList)Data[i]).Add(double.Parse(rdr[i].ToString())); }
                        Weights.Add(1.0);
                        int count = 0;
                        while (rdr.Read())
                        {
                            for (int i = 0; i < Dimension; i++) { ((ArrayList)Data[i]).Add(double.Parse(rdr[i].ToString())); }
                            //SqlContext.Pipe.Send("------- " + count.ToString());

                            for (int i = 0; i < Dimension; i++)
                            {
                                //SqlContext.Pipe.Send("defergerger 2.4 " + rdr[i].ToString() );
                                //SqlContext.Pipe.Send("defergerger 2.5 " + ((double)((ArrayList)Data[i])[count]).ToString());
                            }
                            Weights.Add(1.0);
                            //SqlContext.Pipe.Send("defergerger 2.5 " + ((double)Weights[count]).ToString());
                            count++;
                        }
                        //SqlContext.Pipe.Send("defergerger 2.5");
                    }
                    else
                    {
                        //SqlContext.Pipe.Send("defergerger 3.5");
                        for (int i = 0; i < Dimension; i++) { ((ArrayList)Data[i]).Add(double.Parse(rdr[i].ToString())); }
                        Weights.Add(double.Parse(rdr[Dimension].ToString()));
                        while (rdr.Read())
                        {
                            for (int i = 0; i < Dimension; i++) { ((ArrayList)Data[i]).Add(double.Parse(rdr[i].ToString())); }
                            Weights.Add(double.Parse(rdr[Dimension].ToString()));
                        }
                    }
                    //SqlContext.Pipe.Send("defergerger 4");
                    rdr.Close();
                    double[,] Hist = HistogramNDim(Dimension, Data, Weights, String_X1, String_X2, String_NumCell, HasOpenUpperBound);
                    return Hist;
                     
                }
                catch
                {
                    throw;
                }
                finally
                {
                    connection.Close();
                }
            }
        }


        public double[,] HistogramNDim(int Dimension, ArrayList Data, ArrayList Weights, string String_X1, string String_X2, string String_NumCell, bool HasOpenUpperBound)
        {
            return HistogramNDim(Dimension, Data, Weights, String_X1, String_X2, String_NumCell, HasOpenUpperBound, int.MaxValue);
        }

        public double[,] HistogramNDim(int Dimension, ArrayList Data, ArrayList Weights, string String_X1, string String_X2, string String_NumCell, bool HasOpenUpperBound, int MaxNumCells)
        {
            if (Dimension < 1)
                throw new System.Exception("Invalid paramter(s) in MathFunctions.HistogramNDim. The value of Dimension should be greater or equal than 1.");

            //-------------------------------------------------------------------------------------------------------
            int LengthData = Weights.Count;
            ArrayList X1a = new ArrayList();
            ArrayList X2a = new ArrayList();
            ArrayList NumCella = new ArrayList();

            char[] FieldSeparator = { ',', ' ', '\t' };

            string[] columns1;
            columns1 = String_X1.Split(FieldSeparator, Dimension, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < columns1.GetLength(0); i++)
            {
                if (columns1[i] != "" && columns1[i] != "," && columns1[i] != " ") { X1a.Add(double.Parse((columns1[i].ToString()))); }
            }
            columns1 = String_X2.Split(FieldSeparator, Dimension, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < columns1.GetLength(0); i++)
            {
                if (columns1[i] != "" && columns1[i] != "," && columns1[i] != " ") { X2a.Add(double.Parse((columns1[i].ToString()))); };
            }
            columns1 = String_NumCell.Split(FieldSeparator, Dimension, StringSplitOptions.RemoveEmptyEntries);

            for (int i = 0; i < columns1.GetLength(0); i++)
            {
                if (columns1[i] != "" && columns1[i] != "," && columns1[i] != " ") { NumCella.Add(int.Parse((columns1[i].ToString()))); };
            }

            double[] X1 = new double[Dimension]; X1a.CopyTo(X1);
            double[] X2 = new double[Dimension]; X2a.CopyTo(X2);
            int[] NumCell = new int[Dimension]; NumCella.CopyTo(NumCell);

            //-------------------------------------------------------------------------------------------------------
            if (X1.GetLength(0) != Dimension || X2.GetLength(0) != Dimension || NumCell.GetLength(0) != Dimension)
                throw new System.Exception("Invalid paramter(s) in MathFunctions.HistogramNDim. Number of elements in String_X1,String_X2 and NumCell should be equal to Dimension.");
            int numcell = 1;
            for (int i = 0; i < Dimension; i++)
            {
                if (X1[i] >= X2[i] || NumCell[i] <= 0)
                    throw new System.Exception("Invalid paramter(s) in MathFunctions.HistogramNDim. Each elements of String_X1 should be less than its counterpart in String_X2; and each element in NumCell should be greater than zero.");
                numcell = numcell * NumCell[i];
            }
            if (numcell > MaxNumCells)
                throw new System.Exception("The total number of cells in the N-dimensional grid should be equal or less than " + MaxNumCells.ToString().Trim());

            //-------------------------------------------------------------------------------------------------------
            int[] BlockSize = new int[Dimension];
            double[] IntervalX = new double[Dimension];
            double[] DeltaX = new double[Dimension];
            int[] IndexArray = new int[Dimension];
            //-------------------------------------------------------------------------------------------------------
            int LengthHistogram = 1;

            for (int i = 0; i < Dimension; i++)
            {
                DeltaX[i] = (double)((double)X2[i] - (double)X1[i]) / NumCell[i];
                IntervalX[i] = DeltaX[i] * NumCell[i];
                LengthHistogram = LengthHistogram * ((int)NumCell[i]);
            }
            //-------------------------------------------------------------------------------------------------------
            BlockSize[0] = LengthHistogram / ((int)NumCell[0]);
            for (int i = 1; i < Dimension; i++)
            {
                BlockSize[i] = (int)(BlockSize[i - 1] / ((int)NumCell[i]));
            }
            //-------------------------------------------------------------------------------------------------------
            double[,] Histogram = new double[LengthHistogram, (Dimension + 2)]; //0:(Dimension-1)=paramters ; Dimension=histogram's height, Dimension+1=histogram's height normalized
            //-------------------------------------------------------------------------------------------------------
            double[] Point = new double[Dimension];
            int IndexHistogram = 0;
            int RejectFlag = 0;
            for (int i = 0; i < LengthHistogram; i++)
            {
                IndexArray[0] = (int)(i / (BlockSize[0]));
                for (int j = 1; j < Dimension; j++)
                {
                    int sum = 0;
                    for (int k = 0; k < j; k++)
                    {
                        sum = sum + IndexArray[k] * BlockSize[k];
                    }
                    if (sum == 0 || i <= sum) { IndexArray[j] = 0; }
                    if (i > sum) { IndexArray[j] = (int)((i - sum) / BlockSize[j]); }
                }
                for (int j2 = 0; j2 < Dimension; j2++)
                {
                    if (IndexArray[j2] < (NumCell[j2]))
                    {
                        Histogram[i, j2] = X1[j2] + (IndexArray[j2] + 0.5) * DeltaX[j2];
                    }
                }
            }
            //-------------------------------------------------------------------------------------------------------
            double SumWeight = 0;
            if (HasOpenUpperBound == true)
            {
                for (int i = 0; i < LengthData; i++)
                {
                    RejectFlag = 0;
                    for (int j = 0; j < Dimension; j++)
                    {
                        Point[j] = ((double)(((ArrayList)Data[j])[i]));
                        if (Point[j] < X1[j] || Point[j] >= X2[j])
                        { RejectFlag = 1; j = Dimension; }
                    }
                    if (RejectFlag == 0)
                    {
                        IndexHistogram = 0;
                        for (int j = 0; j < Dimension; j++)
                        {
                            IndexArray[j] = (int)Math.Floor(Round((Point[j] - X1[j]) / DeltaX[j], digits));
                            IndexHistogram = IndexHistogram + IndexArray[j] * BlockSize[j];
                        }
                        Histogram[IndexHistogram, Dimension] = Histogram[IndexHistogram, Dimension] + (double)Weights[i];
                        Histogram[IndexHistogram, Dimension + 1] = Histogram[IndexHistogram, Dimension + 1] + 1.0;
                        SumWeight = SumWeight + (double)Weights[i];
                    }
                }
            }
            else
            {
                for (int i = 0; i < LengthData; i++)
                {
                    RejectFlag = 0;
                    for (int j = 0; j < Dimension; j++)
                    {
                        Point[j] = ((double)(((ArrayList)Data[j])[i]));
                        if (Point[j] <= X1[j] || Point[j] > X2[j])
                        { RejectFlag = 1; j = Dimension; }
                    }
                    if (RejectFlag == 0)
                    {
                        IndexHistogram = 0;
                        for (int j = 0; j < Dimension; j++)
                        {
                            IndexArray[j] = (int)Math.Ceiling(Round((Point[j] - X1[j]) / DeltaX[j], digits)) - 1; ;// (int)Math.Floor(NumCell[j] * (Point[j] - X1[j]) / IntervalX[j]);
                            IndexHistogram = IndexHistogram + IndexArray[j] * BlockSize[j];
                        }
                        Histogram[IndexHistogram, Dimension] = Histogram[IndexHistogram, Dimension] + (double)Weights[i];
                        Histogram[IndexHistogram, Dimension + 1] = Histogram[IndexHistogram, Dimension + 1] + 1.0;
                        SumWeight = SumWeight + (double)Weights[i];
                    }
                }
            }
            double VolumeNdim = 1;
            for (int i = 0; i < Dimension; i++)
            {
                VolumeNdim = VolumeNdim * IntervalX[i] / NumCell[i];
            }
            double NormalizationFactor = VolumeNdim * SumWeight;
            for (int i = 0; i < LengthHistogram; i++)
            {
                Histogram[i, Dimension] = Histogram[i, Dimension] / NormalizationFactor;
            }

            return Histogram;
        }



        public double[,] HistogramNDim(int Dimension, SqlConnection connection, SqlCommand cmd, string String_X1, string String_X2, string String_NumCell, bool HasOpenUpperBound, int MaxNumCells)
        {
            connection.Open();
            SqlDataReader rdr = cmd.ExecuteReader();

            rdr.Read();
            //string ColName = rdr.GetName(0);
            string QueryResultString = rdr[0].ToString();

            if (rdr.FieldCount > (Dimension + 1) || rdr.FieldCount == 0 || rdr.FieldCount < Dimension)
                throw new System.Exception("Incorrect number of query result columns or Dimension value in SQLmath.SQLMathFunctions.HistogramNDimFromQuery. Number of Columns returned by QueryToTable (=" + rdr.FieldCount.ToString().Trim() + ") should be equal to either " + Dimension.ToString().Trim() + " or " + ((int)(Dimension + 1)).ToString().Trim());

            if (Dimension < 1)
                throw new System.Exception("Invalid paramter(s) in MathFunctions.HistogramNDim. The value of Dimension should be greater or equal than 1.");

            //string ColName = rdr.GetName(0);


            bool ExistWeights;
            if (rdr.FieldCount == Dimension)
                ExistWeights = false;
            else
                ExistWeights = true;

            //-------------------------------------------------------------------------------------------------------
            //int LengthData = Weights.Count;
            ArrayList X1a = new ArrayList();
            ArrayList X2a = new ArrayList();
            ArrayList NumCella = new ArrayList();

            char[] FieldSeparator = { ',', ' ', '\t' };

            string[] columns1;
            columns1 = String_X1.Split(FieldSeparator, Dimension, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < columns1.GetLength(0); i++)
            {
                if (columns1[i] != "" && columns1[i] != "," && columns1[i] != " ") { X1a.Add(double.Parse((columns1[i].ToString()))); }
            }

            columns1 = String_X2.Split(FieldSeparator, Dimension, StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < columns1.GetLength(0); i++)
            {
                if (columns1[i] != "" && columns1[i] != "," && columns1[i] != " ") { X2a.Add(double.Parse((columns1[i].ToString()))); };
            }
            columns1 = String_NumCell.Split(FieldSeparator, Dimension, StringSplitOptions.RemoveEmptyEntries);

            for (int i = 0; i < columns1.GetLength(0); i++)
            {
                if (columns1[i] != "" && columns1[i] != "," && columns1[i] != " ") { NumCella.Add(int.Parse((columns1[i].ToString()))); };
            }

            double[] X1 = new double[Dimension]; X1a.CopyTo(X1);
            double[] X2 = new double[Dimension]; X2a.CopyTo(X2);
            int[] NumCell = new int[Dimension]; NumCella.CopyTo(NumCell);

            //-------------------------------------------------------------------------------------------------------
            if (X1.GetLength(0) != Dimension || X2.GetLength(0) != Dimension || NumCell.GetLength(0) != Dimension)
                throw new System.Exception("Invalid paramter(s) in MathFunctions.HistogramNDim. Number of elements in String_X1,String_X2 and NumCell should be equal to Dimension.");
            int numcell = 1;
            for (int i = 0; i < Dimension; i++)
            {
                if (X1[i] >= X2[i] || NumCell[i] <= 0)
                    throw new System.Exception("Invalid paramter(s) in MathFunctions.HistogramNDim. Each elements of String_X1 should be less than its counterpart in String_X2; and each element in NumCell should be greater than zero.");
                numcell = numcell * NumCell[i];
            }
            if (numcell > MaxNumCells)
                throw new System.Exception("The total number of cells in the N-dimensional grid should be equal or less than " + MaxNumCells.ToString().Trim());

            //-------------------------------------------------------------------------------------------------------
            int[] BlockSize = new int[Dimension];
            double[] IntervalX = new double[Dimension];
            double[] DeltaX = new double[Dimension];
            int[] IndexArray = new int[Dimension];
            //-------------------------------------------------------------------------------------------------------
            int LengthHistogram = 1;

            for (int i = 0; i < Dimension; i++)
            {
                DeltaX[i] = (double)((double)X2[i] - (double)X1[i]) / NumCell[i];
                IntervalX[i] = DeltaX[i] * NumCell[i];
                LengthHistogram = LengthHistogram * ((int)NumCell[i]);
            }
            //-------------------------------------------------------------------------------------------------------
            BlockSize[0] = LengthHistogram / ((int)NumCell[0]);
            for (int i = 1; i < Dimension; i++)
            {
                BlockSize[i] = (int)(BlockSize[i - 1] / ((int)NumCell[i]));
            }
            //-------------------------------------------------------------------------------------------------------
            double[,] Histogram = new double[LengthHistogram, (Dimension + 2)]; //0:(Dimension-1)=paramters ; Dimension=histogram's height, Dimension+1=histogram's height normalized
            //-------------------------------------------------------------------------------------------------------
            double[] Point = new double[Dimension];
            int IndexHistogram = 0;
            int RejectFlag = 0;
            for (int i = 0; i < LengthHistogram; i++)
            {
                IndexArray[0] = (int)(i / (BlockSize[0]));
                for (int j = 1; j < Dimension; j++)
                {
                    int sum = 0;
                    for (int k = 0; k < j; k++)
                    {
                        sum = sum + IndexArray[k] * BlockSize[k];
                    }
                    if (sum == 0 || i <= sum) { IndexArray[j] = 0; }
                    if (i > sum) { IndexArray[j] = (int)((i - sum) / BlockSize[j]); }
                }
                for (int j2 = 0; j2 < Dimension; j2++)
                {
                    if (IndexArray[j2] < (NumCell[j2]))
                    {
                        Histogram[i, j2] = X1[j2] + (IndexArray[j2] + 0.5) * DeltaX[j2];
                    }
                }
            }
            //-------------------------------------------------------------------------------------------------------
            double SumWeight = 0;
            bool AreThereMoreRows = true;

            //if (HasOpenUpperBound == true)
            //{
            while (AreThereMoreRows)
            {
                //for (int i = 0; i < LengthData; i++)
                //{
                RejectFlag = 0;
                for (int j = 0; j < Dimension; j++)
                {
                    //Point[j] = ((double)(((ArrayList)Data[j])[i]));
                    Point[j] = double.Parse(rdr[j].ToString());
                    if (Point[j] < X1[j] || Point[j] >= X2[j])
                    { RejectFlag = 1; j = Dimension; }
                }
                if (RejectFlag == 0)
                {
                    IndexHistogram = 0;
                    for (int j = 0; j < Dimension; j++)
                    {
                        if(HasOpenUpperBound)
                            IndexArray[j] = (int)Math.Floor(Round((Point[j] - X1[j]) / DeltaX[j], digits));
                        else
                            IndexArray[j] = (int)Math.Ceiling(Round((Point[j] - X1[j]) / DeltaX[j], digits)) - 1; ;// (int)Math.Floor(NumCell[j] * (Point[j] - X1[j]) / IntervalX[j]);
                        IndexHistogram = IndexHistogram + IndexArray[j] * BlockSize[j];
                    }
                    if (ExistWeights)
                    {
                        Histogram[IndexHistogram, Dimension] = Histogram[IndexHistogram, Dimension] + double.Parse(rdr[Dimension].ToString());
                        SumWeight = SumWeight + double.Parse(rdr[Dimension].ToString());
                    }
                    else
                    {
                        Histogram[IndexHistogram, Dimension] = Histogram[IndexHistogram, Dimension] + 1.0;
                        SumWeight = SumWeight + 1;
                    }
                    Histogram[IndexHistogram, Dimension + 1] = Histogram[IndexHistogram, Dimension + 1] + 1.0;
                }
                if (!rdr.Read())
                    AreThereMoreRows = false;
            }
            rdr.Close();

            double VolumeNdim = 1;
            for (int i = 0; i < Dimension; i++)
            {
                VolumeNdim = VolumeNdim * IntervalX[i] / NumCell[i];
            }
            double NormalizationFactor = VolumeNdim * SumWeight;
            for (int i = 0; i < LengthHistogram; i++)
            {
                Histogram[i, Dimension] = Histogram[i, Dimension] / NormalizationFactor;
            }

            return Histogram;
        }









  
    }





}
