using System.Collections;
using System;
using System.Data;
using System.Data.SqlClient;
using System.Data.SqlTypes;
using Microsoft.SqlServer.Server;

namespace SQLCosmology
{

    public class SQLCosmologicalFunctionsFixedParameters
    {

        /*
        The cosmological parameters are obtained from from a combination of WMAP+BAO+SN results, explained in the following paper:
        "FIVE-YEAR WILKINSON MICROWAVE ANISOTROPY PROBE OBSERVATIONS:COSMOLOGICAL INTERPRETATION" Komatsu et al. (2009) ApJS 180:330-376
		that can be downloaded from:
		http://lambda.gsfc.nasa.gov/product/map/dr3/map_bibliography.cfm
         
         Note: 
         Since OmegaK is very close to zero, the value of OmegaM has been changed slightly, such that we have OmegaM + OmegaL +OmegaK + OmegaR = 1.
         -0.0179 < OmegaK < 0.0081 -> OmegaK~0.0
         h_0=0.705+-0.013
         OmegaL=0.726+-0.015 
         OmegaGamma*h_0^2=2.469e-5 -> OmegaGamma=4.968e-5
         OmegaR = OmegaGamma*(1+0.2271*Neff) where Neff=effective number of neutrinos.
         Neff=4.4+-1.5
         Then 
		 OmegaR = 9.932e-5 ~ 1.0e-4
         OmegaM*h_0^2=0.1358+-0.0037 -> OmegaM= 0.273225693... But OmegaM = 1-0.726-1.0e-4 = 0.2739

        The parameter "w_0" that defines the Dark Energy equation of state at z=0 is set to -1
        */
        //The cosmological paramemters:

        public const double OmegaM = 0.2739;
        public const double OmegaL = 0.7260;
        public const double OmegaR = 0.0001;
        public const double w_0 = -1;
        public const double h_0 = 0.705;

        /*
        //No Radiation density:
        public const double OmegaM = 0.274;
        public const double OmegaK = 0.0;
        public const double OmegaL = 0.726;
        public const double OmegaR = 0;
        public const double w_0 = -1;
        public const double h_0 = 0.705;
        */
        
        public const double TOL_Romberg = 1e-9;
        public const int N_1_Romberg = 5;
        public const double TOL_BracketingWithNeville = 1e-9;
        public const double ExpansionFactor_BracketingWithNeville = 10.0;// to be used in the BracketingWithNeville routine.
        public const int NumVertices_BracketingWithNeville = 5;
        public const int MAXIT_BracketingWithNeville = 5000;


        [SqlFunction(
        Name = "fCosmfZfromTh",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble RedshiftAtHubbleTime(SqlDouble HubbleTIME)
        {
            SqlDouble answer;
            try
            {
                if (HubbleTIME.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtHubbleTime((double)HubbleTIME);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmfZfromDh",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble RedshiftAtHubbleDistance(SqlDouble HubbleDISTANCE)
        {
            SqlDouble answer;
            try
            {
                if (HubbleDISTANCE.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtHubbleDistance((double)HubbleDISTANCE);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }


        
        [SqlFunction(
        Name = "fCosmfZfromDl",    
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise=false
        )]
        public static SqlDouble RedshiftAtLuminosityDistance(SqlDouble LuminosityDistance)
        {
            SqlDouble answer;
            try
            {
                if (LuminosityDistance.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtLuminosityDistance((double)LuminosityDistance);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmfComovVolumeFromDl",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovVolumeAtLumDist(SqlDouble LuminosityDistance)
        {
            SqlDouble answer;
            try
            {
                if (LuminosityDistance.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovVolumeAtLumDist((double)LuminosityDistance);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmfZfromComovVolume",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble RedshiftAtComovVolume(SqlDouble ComovVolume)
        {
            SqlDouble answer;
            try
            {
                if (ComovVolume.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtComovVolume((double)ComovVolume);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }


        public static void FillRowRedshiftAtAngularDiamDist(object item, out SqlDouble z1, out SqlDouble z2)
        {
            double [] row = (double [])item;
            z1 = (SqlDouble)row[0];
            z2 = (SqlDouble)row[1];
        }


        [SqlFunction(
        Name = "fCosmfZfromDa",
        FillRowMethodName = "FillRowRedshiftAtAngularDiamDist",
        TableDefinition = "z1 float, z2 float",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise=false
        )]
        public static IEnumerable RedshiftAtAngularDiamDist(SqlDouble AngularDiamDist)
        {
            ArrayList RedshiftArray = new ArrayList(1);
            try
            {
                if (AngularDiamDist.IsNull)
                    RedshiftArray.Add(new SqlDouble[] {SqlDouble.Null,SqlDouble.Null});
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    RedshiftArray.Add(Cosmo.RedshiftAtAngularDiamDist((double)AngularDiamDist));
                }
            }
            catch
            {
                throw;
            }
            return RedshiftArray;
        }

        [SqlFunction(
        Name = "fCosmfZfromDm",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise=false
        )]
        public static SqlDouble RedshiftAtComovDistTransverse(SqlDouble ComovDistTransverse)
        {
            SqlDouble answer;
            try
            {
                if (ComovDistTransverse.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtComovDistTransverse((double)ComovDistTransverse);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmfZfromDc",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble RedshiftAtComovDistLineOfSight(SqlDouble ComovDistLineOfSight)
        {
            SqlDouble answer;
            try
            {
                if (ComovDistLineOfSight.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtComovDistLineOfSight((double)ComovDistLineOfSight);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmfZfromAgeOfUniverse",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble RedshiftAtAgeOfUniverse(SqlDouble AgeOfUniverse)
        {
            SqlDouble answer;
            try
            {
                if (AgeOfUniverse.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtAgeOfUniverse((double)AgeOfUniverse);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmfZfromLookBackTime",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble RedshiftAtLookBackTime(SqlDouble LookBackTime)
        {
            SqlDouble answer;
            try
            {
                if (LookBackTime.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtLookBackTime((double)LookBackTime);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmfLookBackTime",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble LookBackTime(SqlDouble z)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.LookBackTime((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }


        [SqlFunction(
        Name = "fCosmfTimeInterval",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble TimeInterval(SqlDouble zMin, SqlDouble zMax)
        {
            SqlDouble answer;
            try
            {
                if (zMin.IsNull || zMax.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.TimeInterval((double)zMin, (double)zMax);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }



        [SqlFunction(
        Name = "fCosmfAgeOfUniverse",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble AgeOfUniverse(SqlDouble z)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.AgeOfUniverse((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }


        [SqlFunction(
        Name = "fCosmfHubbleDistance",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise=false
        )]
        public static SqlDouble HubbleDistance(SqlDouble z)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.HubbleDistance((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmfHubbleTime",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble HubbleTime(SqlDouble z)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.HubbleTime((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        
        
        [SqlFunction(
        Name = "fCosmfDl",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble LuminosityDistance(SqlDouble z)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.LuminosityDistance((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmfDc",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise=false
        )]
        public static SqlDouble ComovDistLineOfSight(SqlDouble z)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovDistLineOfSight(0, (double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }


        [SqlFunction(
        Name = "fCosmfDcInterval",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovDistLineOfSightInterval(SqlDouble Zmin, SqlDouble Zmax)
        {
            SqlDouble answer;
            try
            {
                if (Zmin.IsNull || Zmax.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovDistLineOfSight((double)Zmin, (double)Zmax);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmfComovDist2ObjectsRaDec",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovDistBetween2objectsRaDec(SqlDouble Redshift1, SqlDouble RA1, SqlDouble DEC1, SqlDouble Redshift2, SqlDouble RA2, SqlDouble DEC2)
        {
            SqlDouble answer;
            try
            {
                if (Redshift1.IsNull || Redshift2.IsNull || RA1.IsNull || DEC1.IsNull || RA2.IsNull || DEC2.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovDistBetween2objectsRaDec((double)Redshift1, (double)RA1, (double)DEC1, (double)Redshift2, (double)RA2, (double)DEC2);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmfComovDist2ObjectsXYZ",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovDistBetween2objectsXyz(SqlDouble Redshift1, SqlDouble x1, SqlDouble y1, SqlDouble z1, SqlDouble Redshift2, SqlDouble x2, SqlDouble y2, SqlDouble z2)
        {
            SqlDouble answer;
            try
            {
                if (Redshift1.IsNull || Redshift2.IsNull || x1.IsNull || y1.IsNull || z1.IsNull || x2.IsNull || y2.IsNull || z2.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovDistBetween2objectsXyz((double)Redshift1, (double)x1, (double)y1, (double)z1, (double)Redshift2, (double)x2, (double)y2, (double)z2);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmfComovDist2Objects",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovDistBetween2objects(SqlDouble Redshift1, SqlDouble Redshift2, SqlDouble AngularSeparation)
        {
            SqlDouble answer;
            try
            {
                if (Redshift1.IsNull || Redshift2.IsNull || AngularSeparation.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovDistBetween2objects((double)Redshift1, (double)Redshift2, (double)AngularSeparation);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fMathAngSepXYZ",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble AngularSeparationXyz(SqlDouble x1, SqlDouble y1, SqlDouble z1, SqlDouble x2, SqlDouble y2, SqlDouble z2)
        {
            SqlDouble answer;
            try
            {
                if (x1.IsNull || y1.IsNull || z1.IsNull || x2.IsNull || y2.IsNull || z2.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Mathematics.MathFunctions MATH = new Mathematics.MathFunctions();
                    answer = MATH.AngularSeparationXyz((double)x1, (double)y1, (double)z1, (double)x2, (double)y2, (double)z2);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }


        [SqlFunction(
        Name = "fMathAngSepRADEC",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble AngularSeparationRaDec(SqlDouble RA_1, SqlDouble DEC_1, SqlDouble RA_2, SqlDouble DEC_2)
        {
            SqlDouble answer;
            try
            {
                if (RA_1.IsNull || DEC_1.IsNull || RA_2.IsNull || DEC_2.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Mathematics.MathFunctions MATH = new Mathematics.MathFunctions();
                    answer = MATH.AngularSeparationRaDec((double)RA_1, (double)DEC_1, (double)RA_2, (double)DEC_2);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }


        [SqlFunction(
        Name = "fCosmfDa",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovDistTransverse(SqlDouble z)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovDistTransverse((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer; 
        }

        [SqlFunction(
        Name = "fCosmfDm",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble AngularDiameterDist(SqlDouble z)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.AngularDiameterDist((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer; 
        }

        [SqlFunction(
        Name = "fCosmfComovingVolume",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovVolume(SqlDouble z)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovVolume((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer; 
        }

        [SqlFunction(
        Name = "fCosmfAbsMagFromLumDist",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble AbsMagFromLumDist(SqlDouble m, SqlDouble LuminosityDistance)
        {
            SqlDouble answer;
            try
            {
                if (LuminosityDistance.IsNull || m.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.AbsMagFromLumDist((double)m, (double)LuminosityDistance, 0.0);
                }
            }
            catch
            {
                throw;
            }
            return answer; 
        }

        [SqlFunction(
        Name = "fCosmfAbsMag",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble AbsMag(SqlDouble m, SqlDouble z)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull || m.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.AbsMag((double)m, (double)z, 0.0);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        public static void FillRowGridAllCosmoQuantitiesIEnumerable(object item, out SqlDouble z, out SqlDouble Dc, out SqlDouble Dm, out SqlDouble Da, out SqlDouble Dl, out SqlDouble ComVol, out SqlDouble LookBackTime, out SqlDouble AgeOfUniverse)
        {
            double[] row = (double[])(item);
            z = SqlDouble.Parse(row[0].ToString());
            Dc = SqlDouble.Parse(row[1].ToString());
            Dm = SqlDouble.Parse(row[2].ToString());
            Da = SqlDouble.Parse(row[3].ToString());
            Dl = SqlDouble.Parse(row[4].ToString());
            ComVol = SqlDouble.Parse(row[5].ToString());
            LookBackTime= SqlDouble.Parse(row[6].ToString());
            AgeOfUniverse= SqlDouble.Parse(row[7].ToString());
        }

        [SqlFunction(
        Name = "fCosmfAllQuantities",
        FillRowMethodName = "FillRowGridAllCosmoQuantitiesIEnumerable",
        TableDefinition = "z float, Dc float, Dm float, Da float, Dl float, ComovingVolume float, LookBackTime float, AgeOfUniverse float",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static IEnumerable GridAllCosmoQuantitiesIEnumerable(SqlDouble zMin, SqlDouble zMax, int NumCell)
        {

            if (zMin.IsNull || zMax.IsNull)
                yield return new SqlDouble[] { SqlDouble.Null, SqlDouble.Null, SqlDouble.Null, SqlDouble.Null, SqlDouble.Null, SqlDouble.Null, SqlDouble.Null, SqlDouble.Null };
            else
            {
                Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                foreach (double[] row in Cosmo.GridAllCosmoQuantities((double)zMin, (double)zMax, NumCell))
                {
                    yield return row;
                }
            }
        }


        static public string Revision
        {
            get { return "$Revision: 1.0 $"; }
        }

       

        /* Revision History

                $Log: msdevcvs.aspx,v $
                Revision 1.1  2008/06/20 10:59:55  mtaghiza
                 Created Revision and History for this class.
        */
    }

}




    